/*
	CS148 Assignment 7 Starter Code
	Wavelet Compression
*/

#include <stdlib.h>
#include <math.h>

// May need to change this include to customize to your environment
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "st.h"

// Current size of the OpenGL window
int windowSizeX;
int windowSizeY;

enum DisplayMode {DISPLAY_SRCIMG, DISPLAY_QHAAR, DISPLAY_RECONSTRUCTED, DISPLAY_INVALID};
DisplayMode displayMode;

float qfactor = 0.f;

STScene scene;  // The scene graph
STTexture* squareTex;

// image information
int img_width, img_height;
// number of Haar transform steps to apply
int transform_steps;

// this is the input image we want to compress
STImage srcImg;
// this is the floating point version of the input image
STHDRImage haarImg;
// this is the quantized version of the Haar wavelet transformed image
STImage qhaarImg;
// this is floating point image working space for inverting the image
STHDRImage inverseImg;
// this is the image reconstructed
STImage reconstructedImg;
// this is the image showing the error between the source and reconstructed images
STImage errorImg;

// these are temporary arrays used for performing transformations
STColor* tmp_input;
STColor* tmp_output;


/* A very simple recursive, integer base 2 log function. */
int log2(int n) {
	if (n == 1 || n == 0) return 0;
	return 1 + log2(n / 2);
}

/*
	Performs a 1D Haar transform on the input data and places the result in
	output.  Both the input and output are length elements long.  The
	averages should be put in the first half of output and the details
	in the second half.
*/
void forward_haar_1d(STColor* input, STColor* output, int length) {

}

/*
	Performs a 1D inverse Haar transform on the input data and places the 
	result in output.  Both the input and output are length elements long.  
	The	averages are in the first half of input and the details
	in the second half.
*/
void inverse_haar_1d(STColor* input, STColor* output, int length) {

}


/* Utility function to copy a row of an image to the temporary input array. */
void read_input_row(STHDRImage& src_img, int row, int length, STColor* tmp_input) {
	for(int i = 0; i < length; i++)
		tmp_input[i] = src_img(i, row);
}
/* Utility function to copy a column of an image to the temporary input array. */
void read_input_col(STHDRImage& src_img, int col, int length, STColor* tmp_input) {
	for(int i = 0; i < length; i++)
		tmp_input[i] = src_img(col, i);
}
/* Utility function to store the temporary output array to an image row. */
void write_output_row(STHDRImage& dest_img, int row, int length, STColor* tmp_output) {
	for(int i = 0; i < length; i++)
		dest_img(i, row) = tmp_output[i];
}
/* Utility function to store the temporary output array to an image column. */
void write_output_col(STHDRImage& dest_img, int col, int length, STColor* tmp_output) {
	for(int i = 0; i < length; i++)
		dest_img(col, i) = tmp_output[i];
}

/*
	Performs a forward Haar transform on an image.  This method uses the non-standard
	wavelet decomposition.
*/
void forward_haar_2d(STHDRImage& src, STHDRImage& dest, int transform_steps) {

}

/*
	Performs an inverse Haar transform on an image.  This method uses the non-standard
	wavelet decomposition.
*/
void inverse_haar_2d(STHDRImage& src, STHDRImage& dest, int transform_steps) {

}

int get_wavelet_level(int c, int r) {
	int max_cr = (c > r) ? c : r;
	return (max_cr == 0) ? 0 : log2(max_cr) + 1;
}

/*
	This routine quantizes a transformed matrix so.  This is where most
	of the loss in lossy compression happens.  Choosing the right values
	to quantize by is very important.
*/
void quantize(STHDRImage& src, STImage& dest) {

}

/*
	This routine inverts the quantization step.
*/
void unquantize(STImage& src, STHDRImage& dest) {

}

void compress_image() {
	// Copy the src image into the dest image
	for(int r = 0; r < img_height; r++) {
		for(int c = 0; c < img_width; c++) {
			haarImg(c,r) = STColor(srcImg(c,r));
		}
	}

	forward_haar_2d(haarImg, haarImg, transform_steps);
	quantize(haarImg, qhaarImg);
}

void decompress_image() {
	unquantize(qhaarImg, inverseImg);
	inverse_haar_2d(inverseImg, inverseImg, transform_steps);

	for(int r = 0; r < img_height; r++) {
		for(int c = 0; c < img_width; c++) {
			reconstructedImg(c,r) = inverseImg(c,r).ToPixel();
		}
	}
}

float entropy(STImage& img) {
	int counts[256];
	memset(counts, 0, sizeof(int)*255);

	for(int r = 0; r < img_height; r++) {
		for(int c = 0; c < img_width; c++) {
			counts[img(c,r).r]++;
			counts[img(c,r).g]++;
			counts[img(c,r).b]++;
		}
	}
	int total = img_width*img_height*3;

	float ent = 0.f;
	for(int i = 0; i < 255; i++) {
		float p_x = float(counts[i]) / total;
		if (p_x > 0.f)
			ent += -p_x * log(p_x) / log(2.f);
	}
	return ent;
}

/*
	This routine updated the texture used on the quad with one of the images
	in the application.  The ability to display the original source, matte,
	and foreground images is useful for debugging
*/
static void updateTexture() {

	switch (displayMode) {
		case DISPLAY_SRCIMG: 
			squareTex->Upload(srcImg, false);
			break;
		case DISPLAY_QHAAR: 
			squareTex->Upload(qhaarImg, false);
			break;
		case DISPLAY_RECONSTRUCTED: 
			squareTex->Upload(reconstructedImg, false);
			break;
	}

	glutPostRedisplay();
}

/* Helper function.  This just performs all the steps we need to
do each time we change a compression parameter.
*/
void compress() {
	compress_image();
	decompress_image();
	printf("Source entropy: %f\n", entropy(srcImg));
	printf("Transformed entropy: %f\n", entropy(qhaarImg));
	updateTexture();
}

/*
	This routine saves one of the images in the application.
	Useful for debugging
*/
static void saveCurrentImg() {

	switch (displayMode) {
		case DISPLAY_SRCIMG: 			
			srcImg.Save("assign7_src.png");
			break;
		case DISPLAY_QHAAR: 
			qhaarImg.Save("assign7_compressed.png");
			break;
		case DISPLAY_RECONSTRUCTED: 
			reconstructedImg.Save("assign7_reconstructed.png");
			break;
	}
}

static void display() {

	glClearColor(.3f, .3f, .3f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	scene.Draw();

	glutSwapBuffers();
}

static void keyboard(unsigned char key, int x, int y) {

	switch (key) {

		case char(27):
		case 'q':
		case 'Q':
			exit(0);

		// save currently displayed image
		case 's':
			saveCurrentImg();
			break;

		// toggle display mode forward
		case '+':
		case '=':
			displayMode = (DisplayMode)(displayMode + 1);
			if (displayMode == DISPLAY_INVALID)
				displayMode = DISPLAY_SRCIMG;
			updateTexture();
			break;

		// toggle display mode backward
		case '-':
		case '_':

			if (displayMode == 0)
				displayMode = DISPLAY_INVALID;
			displayMode = (DisplayMode)(displayMode - 1);
			if (displayMode == DISPLAY_INVALID)
				displayMode = DISPLAY_SRCIMG;
			updateTexture();
			break;
	}

	glutPostRedisplay();
}

static void special(int key, int x, int y) {
	switch(key) {
		case GLUT_KEY_UP:
			qfactor += .1f;
			break;
		case GLUT_KEY_DOWN:
			qfactor -= .1f;
			break;
	}

	// qfactor should be in the range (0, 1]
	//if (qfactor <= 0.f) qfactor = 1E-5f;
	//if (qfactor > 1.f) qfactor = 1.f;

	compress();
}

void reshape(int w, int h) {
	windowSizeX = w;
	windowSizeY = h;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

	gluOrtho2D(0., 1., 0., 1.);
	glViewport(0, 0, w, h);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

}


/*
	Sets up the scene (a single shape with a texture map on it) and
	allocates all the image objects.  

*/
static void setup(const char* srcFile) {
	srcImg.Load(srcFile);

	img_width = srcImg.GetWidth();
	img_height = srcImg.GetHeight();
	int lw = log2(img_width), lh = log2(img_height);
	
	// check that the dimensions are powers of 2
	if ( (1 << lw) != img_width || (1 << lh) != img_height) {
		printf("Dimensions must be powers of 2...\n");
		exit(-1);
	}

	// determine the number of transform steps
	transform_steps = (lw < lh) ? lw : lh;
	int max_len = 1 << (lw > lh ? lw : lh);

	// resize all images to be the same size as the source image
	haarImg.Resize(img_width, img_height);
	qhaarImg.Resize(img_width, img_height);
	inverseImg.Resize(img_width, img_height);
	reconstructedImg.Resize(img_width, img_height);
	errorImg.Resize(img_width, img_height);

	// allocate working space for 1D transforms
	tmp_input = new STColor[max_len];
	tmp_output = new STColor[max_len];

	// Now build the scene graph

	STSceneNode* root = scene.GetRoot();
	STShape* square = STShape::CreateRect(1.f, 1.f);
	square->SetColor(STColor(1.f, 1.f, 1.f));

	squareTex = new STTexture();
	squareTex->Upload(srcImg, false);
	square->SetTexture(squareTex);
	
	STSceneNode* node = root->PlaceObj(square);
	node->SetTransform( STTransform::Translate(.5f, .5f, 0.f) );
}

void usage(char* argv0) {
	fprintf(stderr, "Usage: %s imgfile\n", argv0);
}

int main(int argc, char** argv) {

	std::string srcFilename;

	if (argc < 2) {
		usage(argv[0]);
		exit(1);
	}

	srcFilename = std::string( argv[1] );

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowPosition(20,20);
	glutInitWindowSize(512, 512);
	glutCreateWindow("CS148 - Assignment 7");

	setup(srcFilename.c_str());
	compress();

	glutReshapeWindow(srcImg.GetWidth(), srcImg.GetHeight());

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(special);

	glutMainLoop();

	return 0;
}